# Text::FreeBSD::ToXML
#
# Copyright 2003 by Peter Pentchev <roam@ringlet.net>
#
# This library is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself. 
#
# $Ringlet: perl/sys/periodicxml/Text/FreeBSD/ToXML/ToXML.pm,v 1.10 2003/07/05 05:02:23 roam Exp $

package Text::FreeBSD::ToXML;

use 5.005;
use strict;
use vars qw($VERSION @ISA);

use Date::Manip;
use XML::NETWriter;

@ISA = qw();

$VERSION = '0.00_01';
$VERSION = eval $VERSION;  # see L<perlmodstyle>


# Preloaded methods go here.

my %repfunc_init = (
    'Disk status:'				=> \&repfunc_status_disks,
    'Network interface status:'			=> \&repfunc_status_iface,
    'Doing login accounting:'			=> \&repfunc_login_acct,
    'Checking for uids of 0:'			=> \&repfunc_uid_zero,
    'Check for files with an unknown user or group:'	=> \&repfunc_orphans,
    'Check for out of date packages:'		=> \&repfunc_packages,
    "Last dump(s) done (Dump '>' file systems):"	=> 	\&repfunc_dumps,
    'Checking setuid files and devices:'	=> \&repfunc_setuid,
    'HOST setuid diffs:'			=> \&repfunc_setuid,
    'HOST kernel log messages:'			=> \&repfunc_kernlog,
	      );

my %repsimple_init = (
    'Removing stale files from /var/preserve:'	=> 'clean-preserve',
    'Cleaning out old system announcements:'	=> 'clean-announce',
    'Removing stale files from /var/rwho:'	=> 'clean-rwho',
    'Backup passwd and group files:'		=> 'backup-passwd',
    'Verifying group file syntax:'		=> 'verify-group',
    'Security check:'				=> 'security-check',
    'Checking for rejected mail hosts:'		=> 'status-mail-reject',
    'HOST denied packets:'			=> 'denied-packets',
    'HOST kernel log messages:'			=> 'kernel-log',
    'HOST login failures:'			=> 'login-fail',
    'HOST refused connections:'			=> 'refused-conn',
    'Cleaning up kernel database files:'	=> 'clean-kernel',
    'Rebuilding locate database:'		=> 'rebuild-locate',
    'Rebuilding whatis database:'		=> 'rebuild-whatis',
		);

my %repfunc_default = %repfunc_init;

my %repsimple_default = %repsimple_init;

sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {};

	$self->{'REPFUNC'} = {};
	%{$self->{'REPFUNC'}} = %repfunc_default;
	$self->{'REPSIMPLE'} = {};
	%{$self->{'REPSIMPLE'}} = %repsimple_default;
	$self->{'SUBJECT'} = '(no report processed yet)';
	bless($self, $class);
	return $self;
}

sub rep_parse_date($) {
	my ($datestr) = @_;
	my ($date, $pdate, $ptime);

	$date = ParseDate($datestr);
	if ($date =~ /^(\d{8})(.*)$/) {
		$pdate = $1;
		$date = $2;
	} else {
		$pdate = "00000000";
	}
	if ($date =~ /^(\d{2}:\d{2}:\d{2})/) {
		$ptime = $1;
	} else {
		$ptime = "00:00:00";
	}
	return ($pdate, $ptime);
}

sub repfunc_simple($ $ $ @) {
	my ($wr, $tag, $head, @lines) = @_;
	my $line;

	$wr->WriteStartElement($tag);
	$wr->WriteAttributeString('name', $head);
	if ($#lines > -1) {
		foreach $line (@lines) {
			$wr->WriteElementString('line', $line);
		}
	}
	$wr->WriteEndElement($tag);
}

sub repfunc_default($ $ @) {
	my ($wr, $head, @lines) = @_;

	repfunc_simple($wr, 'repitem', $head, @lines);
}

sub repfunc_clean_preserve($ $ @) {
	my ($wr, $head, @lines) = @_;

	repfunc_simple($wr, 'repitem-clean-preserve', $head, @lines);
}

sub repfunc_status_disks($ $ @) {
	my ($wr, $head, @lines) = @_;
	my @fields;
	my @sfields = ('dev', 'blocks', 'used', 'avail', 'cap', 'mountpt');
	my ($line, $i, $el, $contents);

	$wr->WriteStartElement('repitem-status-disks');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	shift(@lines);
	while (defined($line = shift(@lines))) {
		@fields = split(/ +/, $line);
		$wr->WriteStartElement('fs');
		$wr->WriteAttributeString($sfields[0], $fields[0]);
		for ($i = 1; $i <= $#fields; $i++) {
			if (($sfields[$i] eq "cap") && ($fields[$i] =~ /(.*)%$/)) {
				$wr->WriteElementString($sfields[$i], $1);
			} else {
				$wr->WriteElementString($sfields[$i], $fields[$i]);
			}
		}
		$wr->WriteEndElement('fs');
	}
	$wr->WriteEndElement('repitem-status-disks');
}

sub repfunc_dumps($ $ @) {
	my ($wr, $head, @lines) = @_;
	my ($line, $dev, $mountpt, $level, $datestr, $parsestr, $date);
	my ($pdate, $ptime);

	$wr->WriteStartElement('repitem-last-dumps');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	while (defined($line = shift(@lines))) {
		next unless ($line =~ /^>\s+(\S+)\s+\(\s*(\S+)\s*\)\s+Last dump:\s+Level\s+(\d+),\s+Date\s+(.*)$/);
		($dev, $mountpt, $level, $datestr) = ($1, $2, $3, $4);
		($pdate, $ptime) = rep_parse_date($datestr);
		$wr->WriteStartElement('fs');
		$wr->WriteAttributeString('dev', $dev);
		$wr->WriteElementString('mountpt', $mountpt);
		$wr->WriteElementString('level', $level);
		$wr->WriteElementString('date', $datestr);
		$wr->WriteElementString('parsed-date', $pdate);
		$wr->WriteElementString('parsed-time', $ptime);
		$wr->WriteEndElement('fs');
	}
	$wr->WriteEndElement('repitem-last-dumps');
}

sub display_setuid_file($ $ $) {
	my ($wr, $which, $rarr) = @_;
	my @hdrs = ('inode', 'mode', 'linkcnt', 'owner', 'group', 'size', 'date');
	my ($i, $pdate, $ptime, @arr);

	@arr = @{$rarr};
	$wr->WriteStartElement($which);
	for ($i = 0; $i <= $#hdrs; $i++) {
		$wr->WriteElementString($hdrs[$i], $arr[$i]);
	}
	($pdate, $ptime) = &rep_parse_date($arr[6]);
	$wr->WriteElementString('parsed-date', $pdate);
	$wr->WriteElementString('parsed-time', $ptime);
	$wr->WriteEndElement($which);
}

sub repfunc_setuid($ $ @) {
	my ($wr, $head, @lines) = @_;
	my ($line, $fname, $type);
	my (%old, %new) = ();

	$wr->WriteStartElement('repitem-setuid');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement('repitem-setuid');
		return;
	}

	while (defined($line = shift(@lines))) {
		next unless ($line =~ /^([><])\s+(\d+)\s+(\S+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+\s+\d+\s+\d+:\d+:\d+\s+\d+)\s+(.*)$/);
#		($dir, $inode, $mode, $linkcnt, $owner, $group, $size, $date, $fname) = ($1, $2, $3, $4, $5, $6, $7, $8, $9);
		if ($1 eq '<') {
			$old{$9} = [ $2, $3, $4, $5, $6, $7, $8 ];
		} else {
			$new{$9} = [ $2, $3, $4, $5, $6, $7, $8 ];
		}
	}
	foreach $fname (sort { $a cmp $b } keys(%old)) {
		if (exists($new{$fname})) {
			$type = 'changed';
		} else {
			$type = 'removed';
		}
		$wr->WriteStartElement('file');
		$wr->WriteAttributeString('fname', $fname);
		$wr->WriteAttributeString('type', $type);
		&display_setuid_file($wr, 'old', $old{$fname});
		&display_setuid_file($wr, 'new', $new{$fname})
		    if (exists($new{$fname}));
		$wr->WriteEndElement('file');
	}
	foreach $fname (sort { $a cmp $b } keys(%new)) {
		next if (exists($old{$fname}));
		$wr->WriteStartElement('file');
		$wr->WriteAttributeString('fname', $fname);
		$wr->WriteAttributeString('type', 'added');
		&display_setuid_file($wr, 'new', $old{$fname});
		$wr->WriteEndElement('file');
	}
	$wr->WriteEndElement('repitem-setuid');
}

sub repfunc_status_iface($ $ @) {
	my ($wr, $head, @lines) = @_;
	my @fields;
	my @sfields = ('name', 'mtu', 'network', 'address', 'ipkts', 'ierrs', 'opkts', 'oerrs', 'coll');
	my ($line, $i);

	$wr->WriteStartElement('repitem-status-iface');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	shift(@lines); # skip the header
	while (defined($line = shift(@lines))) {
		@fields = split(/ +/, $line);
		$wr->WriteStartElement('iface');
		$wr->WriteAttributeString($sfields[0], $fields[0]);
		for ($i = 1; $i <= $#fields; $i++) {
			if ($fields[$i] eq '-') {
				$wr->WriteElementString($sfields[$i], '0');
			} else {
				$wr->WriteElementString($sfields[$i], $fields[$i]);
			}
		}
		$wr->WriteEndElement('iface');
	}
	$wr->WriteEndElement('repitem-status-iface');
}

sub repfunc_login_acct($ $ @) {
	my ($wr, $head, @lines) = @_;
	my @fields;
	my $line;

	$wr->WriteStartElement('repitem-login-acct');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}
	
	@fields = split(/\s+/, shift(@lines));
	if (($#fields != 2) || ($fields[1] ne 'total')) {
		$wr->WriteAttributeString('total', 'INVALID');
		$wr->WriteEndElement();
	}
	$wr->WriteAttributeString('total', $fields[2]);
	while (defined($line = shift(@lines))) {
		next unless ($line =~ /^\s+(\S+)\s+(\S+)$/);
		my ($uname, $time) = ($1, $2);
		$wr->WriteStartElement('user');
		$wr->WriteElementString('username', $uname);
		$wr->WriteElementString('time', $time);
		$wr->WriteEndElement('user');
	}
	$wr->WriteEndElement('repitem-login-acct');
}

sub repfunc_uid_zero($ $ @) {
	my ($wr, $head, @lines) = @_;
	my @fields;
	my $line;

	$wr->WriteStartElement('repitem-uid-zero');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	while (defined($line = shift(@lines))) {
		@fields = split(/\s+/, $line);
		$wr->WriteStartElement('user');
		$wr->WriteElementString('username', $fields[0]);
		$wr->WriteElementString('userid', $fields[1]);
		$wr->WriteEndElement();
	}
	$wr->WriteEndElement('repitem-uid-zero');
}

sub repfunc_orphans($ $ @) {
	my ($wr, $head, @lines) = @_;
	my $line;

	$wr->WriteStartElement('repitem-orphans');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	while (defined($line = shift(@lines))) {
		$line =~ s/^\s*//;
		$wr->WriteElementString('file', $line);
	}
	$wr->WriteEndElement('repitem-orphans');
}

sub repfunc_kernlog($ $ @) {
	my ($wr, $head, @lines) = @_;

	repfunc_simple($wr, 'repitem-kernlog', $head, @lines);
}

sub repfunc_packages($ $ @) {
	my ($wr, $head, @lines) = @_;
	my $line;

	$wr->WriteStartElement('repitem-packages');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	while (defined($line = shift(@lines))) {
		if ($line !~ /^\s*(\S+)\s*(.*?)\s*\((.*)\)$/) {
			$wr->WriteElementString('line', $line);
		} else {
			my ($pkg, $reason, $comment) = ($1, $2, $3);

			$wr->WriteStartElement('package');
			$wr->WriteElementString('name', $pkg);
			$wr->WriteElementString('reason', $reason);
			$wr->WriteElementString('comment', $comment);
			$wr->WriteEndElement('package');
		}
	}
	$wr->WriteEndElement('repitem-packages');
}

sub repinfo_init() {

	return (
	    'subj'	=> '',
	    'text'	=> '',
	    'indent'	=> 0,
	    'pending'	=> 0
	    );
}

sub ProcHeader($ $ $) {
	my ($self, $reflines, $refinfo) = @_;
	my @lines = @{$reflines};
	my $line;

	while (defined($line = shift(@lines))) {
		last if ($line eq "");
		if ($line =~ /^Subject:\s*(.*)$/) {
			my $subj = $1;
			$refinfo->{'subj'} = $subj;
			$self->{'SUBJECT'} = $subj;
			if ($subj =~ /^\s*(\S+)\s+(\S+)\s+(run|check)\s+output$/) {
				my ($hostname, $type) = ($1, $2);

				$refinfo->{'type'} = $type;
				$refinfo->{'hostname'} = $hostname;
			}
		} elsif ($line =~ /^Date:(.*)$/) {
			my $date = $1;
			my $parsed = &ParseDate($date);
			if (defined($parsed) && $parsed ne '') {
				$refinfo->{'parsed-date'} = $parsed;
			}
			if ($date =~ /^\s*(\S+)\s+(\S+)\s+(\S+)/) {
				my ($d1, $d2, $d3) = ($1, $2, $3);

				$refinfo->{'date'} = "$d1 $d2 $d3";
			}
		}
	}
	return @lines;
}

sub ProcBody($ $ $ $) {
	my ($self, $reflines, $refinfo, $wr) = @_;
	my @lines = @{$reflines};
	my ($hname, $head, $line, $match);
	my @replines;

	$hname = $refinfo->{'hostname'};
	while (defined($head = shift(@lines))) {
		next if ($head eq "");
		@replines = ();
		while (defined($line = shift(@lines))) {
			last if ($line eq "");
			push(@replines, $line);
		}
		if ($head =~ /^$hname (.*)/) {
			$match = "HOST $1";
		} else {
			$match = $head;
		}
		next if ($match eq "-- End of ".$refinfo->{type}." output --");
		if (defined($self->{'REPFUNC'}->{$match})) {
			&{$self->{'REPFUNC'}->{$match}}($wr, $head, @replines);
		} elsif (defined($self->{'REPSIMPLE'}->{$match})) {
			&repfunc_simple($wr, "repitem-".$self->{'REPSIMPLE'}->{$match}, $head, @replines);
		} else {
			repfunc_default($wr, $head, @replines);
		}
	}
	return @lines;
}

sub ProcFile($ $ $) {
	my ($self, $fname, $wr) = @_;
	my @lines;
	my %repinfo;

	open(IN, "< $fname") or die("Opening $fname: $!\n");
	@lines = <IN>;
	close(IN);
	map { s/[\r\n]*$//g; } @lines;

	%repinfo = &repinfo_init();
	@lines = $self->ProcHeader(\@lines, \%repinfo);
	
	$wr->WriteStartElement('report');
	if (defined($repinfo{'type'})) {
		$wr->WriteAttributeString('type', $repinfo{'type'});
	} else {
		$wr->WriteAttributeString('type', 'unknown');
	}
	$wr->WriteAttributeString("name", $repinfo{'subj'});
	$wr->WriteAttributeString("hostname", $repinfo{'hostname'});
	$wr->WriteAttributeString("date", $repinfo{'date'});
	if (defined($repinfo{'parsed-date'})) {
		my $full = $repinfo{'parsed-date'};
		my ($date, $time) = (undef, undef);
		my @l = localtime();
		if ($full =~ /^(\d{8})(.*)$/) {
			($date, $full) = ($1, $2);
		} else {
			$date = strftime('%Y%m%d', @l);
		}
		if ($full =~ /^(\d\d:\d\d:\d\d)(.*)/) {
			($time, $full) = ($1, $2);
		} else {
			$time = strftime('%H%M%S', @l);
		}
		$wr->WriteAttributeString('parsed-date', $date);
		$wr->WriteAttributeString('parsed-time', $time);
	}
	@lines = $self->ProcBody(\@lines, \%repinfo, $wr);
	$wr->WriteEndElement('report');
	$wr->WriteEndDocument();
}

sub SetFuncHandler($ $ $) {
	my ($self, $name, $handler) = @_;

	$self->{'REPFUNC'}->{$name} = $handler;
}

sub SetSimpleHandler($ $ $) {
	my ($self, $name, $handler) = @_;

	$self->{'REPSIMPLE'}->{$name} = $handler;
}

sub GetFuncHandler($ $) {
	my ($self, $name) = @_;

	return $self->{'REPFUNC'}->{$name};
}

sub GetSimpleHandler($ $) {
	my ($self, $name) = @_;

	return $self->{'REPSIMPLE'}->{$name};
}

# Make Perl happy (in one way at least ;)
1;
__END__
=head1 NAME

Text::FreeBSD::ToXML - Convert a FreeBSD status report to XML

=head1 SYNOPSIS

  use Text::FreeBSD::ToXML;

  $wr = XML::NETWriter->new();
  $report = Text::FreeBSD::ToXML->new();
  $report->ProcFile('report.txt', $wr);

  $report->SetFuncHandler('Custom heading:', \&process_func);

  sub process_func($writer, $heading, @textlines)

=head1 ABSTRACT

The Text::FreeBSD::ToXML module converts a FreeBSD periodic report
into a structured XML file.  It supports all the items in the reports
generated by the FreeBSD base system's default periodic scripts, and
it is easy to extend its functionality by adding handlers for custom
report items.
  
=head1 DESCRIPTION

The Text::FreeBSD::ToXML module has the following methods available for use:

=over 4

=item B<new>()

This is the constructor for the Text::FreeBSD::ToXML parser.  There are no
configurable arguments.

=item B<GetFuncHandler>($name)

Return the function handling the report chunks starting with the headling
line provided in B<$name>, or B<undef> if none.

=item B<GetSimpleHandler>($name)

Return the XML element name suffix if the value in B<$name> is currently
registered as a simple chunk heading, or B<undef> if it is not.  See the
B<REPORT PARSING> section below for more information.

=item B<ProcFile>($filename, $writer)

The B<ProcFile> method reads a FreeBSD periodic report from a filename and
uses the B<XML::NETWriter> object B<$writer> to create its XML
representation.  The file is processed in chunks separated by empty lines
as explained in the B<REPORT PARSING> section below.

The B<$writer> should already be initialized; the B<ProcFile> method does
not invoke its B<WriteStartDocument>() method.

=item B<SetFuncHandler>($name, $handler)

Set a callback handler for report chunks starting with the heading line
provided in B<$name>.  See the B<REPORT PARSING> section below for more
information.

=item B<SetSimpleHandler>($name, $handler)

Defines a heading for a simple match.  The B<$handler> argument contains
the XML element prefix to be used in the output file.  See the B<REPORT
PARSING> section below for more information.

=back

=head1 REPORT PARSING

The FreeBSD periodic reports are structured in E<chunks> - that is to say,
sequences of lines separated by empty lines.  An example of such a chunk:

    Disk status:
    Filesystem  1K-blocks    Used   Avail Capacity  Mounted on
    /dev/da0s1a    127023   48224   68638    41%    /
    /dev/da0s1d   4798518 4035299  379338    91%    /fs

Chunks are identified by their first line, which serves as a heading.
Each time a new chunk is read, the Text::FreeBSD::ToXML parser attempts
to determine an action appropriate for it.  This is done by examining
the first line - the chunk heading - and comparing it to two sets of
action definitions.

The first set consists of the so-called B<simple chunks>.  Those are
chunks that do not really consist of structured content, but merely a
sequence of "equal" lines.  An example of this type of chunks are the
kernel log messages.  In the periodic report plaintext file:

  straylight.ringlet.net kernel log messages:
  > sio0: 1 more silo overflow (total 3)
  > sio0: 1 more silo overflow (total 4)
  > sio0: 1 more silo overflow (total 5)

In the output XML file, they are represented in the following way:

  <repitem-kernlog name="straylight.ringlet.net kernel log messages">
    <line>sio0: 1 more silo overflow (total 3)</line>
    <line>sio0: 1 more silo overflow (total 4)</line>
    <line>sio0: 1 more silo overflow (total 5)</line>
  </repitem-kernlog>

The only characteristics of a simple chunk are its heading line and the
XML element name.  Thus, the above chunk transformation is defined by
a heading line of 'HOST kernel log messages' (the 'HOST' prefix in any
heading deifnition is automatically replaced with the actual hostname of
the machine generating the report), and an XML element suffix of
'kernlog'.  Such an association could be created by invoking the
B<SetSimpleHandler>() method as follows:

  $parser->SetSimpleHandler('HOST kernel log messages', 'kernlog');

The other kind of chunk actions are those handled by callback functions.
They are defined by invoking the B<SetFuncHandler>() method, and
supplying the heading line and a reference to a callback handler sub.
The handler should accept three arguments: the B<XML::NETWriter> to
write to, the heading line itself (in case one handler is used for
multiple actions), and an array containing all the lines in the chunk
following the heading.  The handler does not have access to any other
part of the input report, either preceding or following the current
chunk.  All that it is supposed to do is examine the array of lines
provided, and use the B<XML::NETWriter> module's methods (e.g.
B<WriteStartElement>(), B<WriteEndElement>(), B<WriteAttributeString>(),
B<WriteElementString>()) to build an XML representation of the data
contained in the current chunk.

=head1 SEE ALSO

B<XML::NETWriter>, B<Text::FreeBSD::XmlToSql>

=head1 AUTHOR

Peter Pentchev, E<lt>roam@ringlet.netE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Peter Pentchev

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
