#!/usr/bin/perl -wT
#
# Convert a daily report e-mail into XML to stdout.
#
# $Ringlet: perl/sys/periodicxml/create/procmailreport.pl,v 1.11 2003/07/05 05:22:58 roam Exp $

use strict;

use Config::Tiny;
use Getopt::Std;
use Mail::NextAddress::Qmail;
use MIME::Entity;
use Text::FreeBSD::ToXML;
use Mail::GnuPG;
use Unix::Syslog qw(:macros :subs);

# Log an error and terminate.
sub disp_err($) {
	my ($msg) = @_;
	
	$msg =~ tr/\r\n/__/;
	$msg .= "\n";
	syslog(LOG_ERR, $msg);
	die($msg);
}

sub repfunc_mailq_qmail($ $ @) {
	my ($wr, $head, @lines) = @_;
	my $line;
	my %msg;
	my ($queued, $notproc) = (0, 0);

	$wr->WriteStartElement('repitem-mailq-qmail');
	$wr->WriteAttributeString('name', $head);
	if ($#lines == -1) {
		$wr->WriteEndElement();
		return;
	}

	while (defined($line = shift(@lines))) {
		if ($line =~ /^messages in queue: (\d+)$/) {
			$queued = $1;
			$wr->WriteElementString('queued', $queued);
		} elsif ($line =~ /^messages in queue but not yet preprocessed: (\d+)$/) {
			$notproc = $1;
			$wr->WriteElementString('not-processed', $notproc);
		} elsif ($line =~ /^(.*?)\s+#(\d+)\s+(\d+)\s+(.*?)\s*$/) {
			($msg{'date'}, $msg{'msgid'}, $msg{'size'}, $msg{'sender'}) = ($1, $2, $3, $4);
		} elsif ($line =~ /^\s*remote\s+(.*)$/) {
			$msg{'rcpt'} = $1;
			$wr->WriteStartElement('message');
			$wr->WriteElementString('id', $msg{'msgid'});
			$wr->WriteElementString('date', $msg{'date'});
			$wr->WriteElementString('size', $msg{'size'});
			$wr->WriteElementString('sender', $msg{'sender'});
			$wr->WriteElementString('rcpt', $msg{'rcpt'});
			$wr->WriteEndElement('message');
		} else {
			$wr->WriteElementString('line', $line);
		}
	}
	$wr->WriteEndElement('repitem-mailq-qmail');
}

sub usage()
{
	die("Usage: procmailreport [-h] [-f cfgfile] [-s cfgsect]\n".
	    "\t-f\tspecify the name of the config file to use;\n".
	    "\t-h\tdisplay this help message and exit;\n".
	    "\t-s\tspecify the name of the config file section to use.\n");
}

sub cfg_read($)
{
	my ($cfg) = @_;
	my ($fname, $section, $cfgobj, $k);

	if (defined($cfg->{'cfg_section'})) {
		$section = $cfg->{'cfg_section'};
	} else {
		$section = 'procmailreport';
	}
	return unless (defined($fname = $cfg->{'cfg_file'}) && -f $fname);
	return unless (defined($cfgobj = Config::Tiny->read($fname)));
	return unless (defined($cfgobj) && defined($cfgobj->{$section}));
	foreach $k (keys(%{$cfgobj->{$section}})) {
		$cfg->{$k} = $cfgobj->{$section}->{$k};
	}
}

MAIN:
{
	my %cfg = (
		'cfg_file'	=> '/etc/procmailreport.conf',
		'cfg_section'	=> 'procmailreport',
		'mail_prefix'	=> 'periodicxml',
		'mail_sender'	=> 'sender@invalid',
		'mail_rcpt'	=> 'postmaster',
		'mail_subject'	=> '[processed] periodic report',
		'log_facility'	=> 'LOG_MAIL',
		'log_progname'	=> 'procmailreport',
		'path'		=> '/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin:/usr/X11R6/bin',
		'pgp_key'	=> undef,
		'pgp_keydir'	=> undef,
		'pgp_passphrase'=> undef,
		'xml_ver'	=> '1.0',
		'xml_charset'	=> 'windows-1251',
	);
	my %opts;
	my ($wr, $p);
	my ($sender, $rcpt, $subj, $top, $gpg, $err);

	getopts('f:hs:', \%opts) or &usage();
	&usage() if defined($opts{'h'});
	$cfg{'cfg_file'} = $opts{'f'} if defined($opts{'f'});
	$cfg{'cfg_section'} = $opts{'s'} if defined($opts{'s'});
	
	&cfg_read(\%cfg);
	
	# Set the search path as early as possible to avoid
	# taint mode problems.
	$ENV{'PATH'} = $cfg{'path'};

	# Initialize the logging subsystem.
	openlog($cfg{'log_progname'}, LOG_PID,
	    eval "Unix::Syslog::$cfg{'log_facility'}");

	# Supply "reasonable defaults"
	($sender, $rcpt) =
	    Mail::NextAddress::Qmail::Guess($cfg{'mail_prefix'});
	$sender = $cfg{'mail_sender'} unless (defined($sender));
	$rcpt = $cfg{'mail_rcpt'} unless (defined($rcpt));

	# Parse the report itself
	eval {
		$wr = new XML::NETWriter();
		$wr->WriteStartDocument($cfg{'xml_ver'}, $cfg{'xml_charset'});
		$p = new Text::FreeBSD::ToXML();
		$p->SetFuncHandler('Mail in local queue:',
		    \&repfunc_mailq_qmail);
		$p->ProcFile('-', $wr);
	};
	&disp_err("Converting: $@") if ($@);
	$subj = $p->{'SUBJECT'};
	if (defined($subj)) {
		$subj = "[processed] $subj";
	} else {
		$subj = $cfg{'mail_subject'};
	}

	# Build the e-mail message...
	$top = MIME::Entity->build(Type => 'multipart/mixed',
	    From => $sender, To => $rcpt, Subject => $subj);
	$top->attach(Type => 'text/xml', Data => $wr->text,
	    Charset => $cfg{'xml_charset'});
	# ...and sign it.
	if (defined($cfg{'pgp_key'})) {
		eval {
			my %pgpargs = ('key' => $cfg{'pgp_key'});
			if (defined($cfg{'pgp_keydir'})) {
				$pgpargs{'keydir'} = $cfg{'pgp_keydir'};
			}
			if (defined($cfg{'pgp_passphrase'})) {
				$pgpargs{'passphrase'} =
				    $cfg{'pgp_passphrase'};
			}
			$gpg = Mail::GnuPG->new(%pgpargs);
			$err = $gpg->mime_sign($top);
		};
		&disp_err("Signing: $@") if ($@);
		if (defined($err) && ($err != 0))
		{
			my @msg = @{$gpg->{last_message}};
			&disp_err("Signing error: $err: ".join('', @msg));
		}
	}

	# Display the message to be handled by procmail and friends.
	$top->print;
}
